from string import ascii_lowercase, digits
from typing import Optional

from surrealist.errors import SurrealRecordIdError

ALPHABET = ascii_lowercase + digits


class RecordId:
    """
    This class is a wrapper for record_id of SurrealDB

    About record_ids: https://surrealdb.com/docs/surrealql/datamodel/ids
    Refer to: https://github.com/kotolex/surrealist?tab=readme-ov-file#using-recordid
    Examples: https://github.com/kotolex/surrealist/blob/master/examples/record_id.py
    """

    def __init__(self, id_: str, table: Optional[str] = None):
        """
        Wrapper for record_id
        :param id_: full record_id like "table:id" or just "id" part of it. In the latter case, the table should be
        specified
        :param table: name of table, you can omit it if id_ is in full form (table:id)
        :raises: SurrealRecordIdError on invalid id or table
        """
        if table is None and ":" not in id_:
            raise SurrealRecordIdError("You need to specify table name or id like 'table:id'")
        if table and ":" in table:
            raise SurrealRecordIdError("Table name should not contain ':'")
        if table and ":" in id_ and table != id_.split(":")[0]:
            table_part = id_.split(":")[0]
            raise SurrealRecordIdError(f"Table name is different from id, we expect {table}, but got {table_part}")
        id_ = id_.replace("`", "").replace("⟨", "").replace("⟩", "")
        self._naive_id = id_ if ":" in id_ else f"{table}:{id_}"
        self._table_part, self._id_part = self._naive_id.split(":")
        self._uid = f"{self._table_part}:⟨{self._id_part}⟩"

    def __repr__(self):
        return f"RecordId('{self._naive_id}')"

    @property
    def id_part(self) -> str:
        """
        Returns id part of record_id, for "table:id" it returns "id"
        """
        return self._id_part

    @property
    def table_part(self) -> str:
        """
        Returns table part of record_id, for "table:id" it returns "table"
        """
        return self._table_part

    @property
    def naive_id(self) -> str:
        """
        Returns naive record_id, without any special braces, e.g. 'table:id'
        """
        return self._naive_id

    def to_valid_string(self) -> str:
        """
        Checks and adds special braces if id is not in simple form(a..zA..Z0-9), otherwise just returns naive_id
        """
        is_complicated_format = any(e not in ALPHABET for e in self._id_part.lower())
        return self.to_uid_string() if is_complicated_format else self._naive_id

    def to_prefixed_string(self) -> str:
        """
        Return record id with r prefix like r'table:id'
        """
        return f"r'{self._naive_id}'"

    def to_uid_string(self) -> str:
        """
        Return record id with special braces for id like article:⟨c332eb25-e408-4396-814f-83a85d556493⟩
        """
        return self._uid

    def to_uid_string_with_backticks(self) -> str:
        """
        Return record id with backticks for id like article:`c332eb25-e408-4396-814f-83a85d556493`
        """
        return self._uid.replace("⟨", "`").replace("⟩", "`")



from typing import Optional

from surrealist.ql.statements.common_statements import CanUseComment
from surrealist.ql.statements.statement import FinishedStatement, Statement


# This module contains classes with duplicate code. This is made specifically for IDE hints

class Concurrently(FinishedStatement):
    """
    Represents CONCURRENTLY statement

    Refer to:
    https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#using-concurrently-clause
    """

    def _clean_str(self):
        return f"{self._statement._clean_str()} CONCURRENTLY"


class CanUseConcurrently:
    def concurrently(self):
        """
        Adds CONCURRENTLY clause to a statement
        """
        return Concurrently(self)


class Unique(FinishedStatement, CanUseComment, CanUseConcurrently):
    """
    Represents UNIQUE statement

    Refer to:
    https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#unique-index
    """

    def _clean_str(self):
        return f"{self._statement._clean_str()} UNIQUE"


class SearchAnalyzer(FinishedStatement, CanUseComment, CanUseConcurrently):
    """
    Represents SEARCH ANALYZER statement

    Refer to:
    https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#full-text-search-index
    """

    def __init__(self, statement: Statement, name: str):
        super().__init__(statement)
        self._name = name
        self._bm25 = None
        self._high = False

    def highlights(self) -> "SearchAnalyzer":
        """
        Used for highlights the matching keywords

        Refer to:
        https://surrealdb.com/docs/surrealdb/reference-guide/full-text-search#highlighting

        :return: SearchAnalyzer object
        """
        self._high = True
        return self

    def bm25(self, k1: Optional[float] = None, b: Optional[float] = None) -> "SearchAnalyzer":
        """
        Use ranking algorithm Best Match 25 (Add BM25 clause to a final statement)
        BM25 [(@k1, @b)]

        Note: you should specify both parameters (k1, b) or none of them. If you specify only one of them,
        the other will be equal to 0.0

        :param k1: optional parameter for BM25
        :param b: optional parameter for BM25
        :return: SearchAnalyzer object
        """
        self._bm25 = (k1, b)
        if (k1, b) != (None, None):
            self._bm25 = (k1 or 0.0, b or 0.0)
        return self

    def _clean_str(self):
        hl = "" if not self._high else " HIGHLIGHTS"
        bm25 = ""
        if self._bm25:
            pair = self._bm25
            bm25 = " BM25" if pair == (None, None) else f" BM25 {pair[0]} {pair[1]}"
        return f"{self._statement._clean_str()} SEARCH ANALYZER {self._name}{bm25}{hl}"


class MTree(FinishedStatement, CanUseComment, CanUseConcurrently):
    """
    Represents MTREE index.
    There is code duplication in this class. This is made specifically for IDE hints

    Refer to:
    https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#vector-search-indexes
    """

    def __init__(self, statement: Statement, dimension_number: int):
        super().__init__(statement)
        self._num = dimension_number
        self._dist = None
        self._capacity = None
        self._type = None

    def distance_euclidean(self) -> "MTree":
        """
        Make index to use Euclidean distance.
        When no function is specified, SurrealDB chooses Euclidean for the default distance function

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/functions/vector#vectordistanceeuclidean

        :return: MTree object
        """
        self._dist = "EUCLIDEAN"
        return self

    def distance_manhattan(self) -> "MTree":
        """
        Make index to use Manhattan distance
        When no function is specified, SurrealDB chooses Euclidean for the default distance function

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/functions/vector#vectordistancemanhattan

        :return: MTree object
        """
        self._dist = "MANHATTAN"
        return self

    def distance_cosine(self) -> "MTree":
        """
        Make index to use Cosine distance
        When no function is specified, SurrealDB chooses Euclidean for the default distance function

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/functions/vector#vectorsimilaritycosine

        :return: MTree object
        """
        self._dist = "COSINE"
        return self

    def distance_minkowski(self) -> "MTree":
        """
        Make index to use Minkowski distance
        When no function is specified, SurrealDB chooses Euclidean for the default distance function

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/functions/vector#vectordistanceminkowski

        :return: MTree object
        """
        self._dist = "MINKOWSKI"
        return self

    def capacity(self, value: int) -> "MTree":
        """
        The CAPACITY clause is used to specify the maximum number of records that can be stored in the index.
        This is useful when you want to limit the number of records stored in the index to optimize performance
        The default value is 40 (on SurrealDB side)

        :param value: integer value for capacity
        :return: MTree object
        """
        self._capacity = value
        return self

    def type_f64(self) -> "MTree":
        """
        The TYPE clause is optional and can be used to specify the data type of the vector.
        Represents 64-bit floating-point numbers (double precision floating-point numbers).

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#types

        :return: MTree object
        """
        self._type = "F64"
        return self

    def type_f32(self) -> "MTree":
        """
        The TYPE clause is optional and can be used to specify the data type of the vector.
        Represents 32-bit floating-point numbers (single precision floating-point numbers).

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#types

        :return: MTree object
        """
        self._type = "F32"
        return self

    def type_i64(self) -> "MTree":
        """
        The TYPE clause is optional and can be used to specify the data type of the vector.
        Represents 64-bit signed integers.

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#types

        :return: MTree object
        """
        self._type = "I64"
        return self

    def type_i32(self) -> "MTree":
        """
        The TYPE clause is optional and can be used to specify the data type of the vector.
        Represents 32-bit signed integers.

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#types

        :return: MTree object
        """
        self._type = "I32"
        return self

    def type_i16(self) -> "MTree":
        """
        The TYPE clause is optional and can be used to specify the data type of the vector.
        Represents 16-bit signed integers.

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#types

        :return: MTree object
        """
        self._type = "I16"
        return self

    def _clean_str(self):
        type_ = f' TYPE {self._type}' if self._type else ''
        dist = f' DIST {self._dist}' if self._dist else ''
        cap = f' CAPACITY {self._capacity}' if self._capacity else ''
        return f"{self._statement._clean_str()} MTREE DIMENSION {self._num}{type_}{dist}{cap}"


class HNSW(FinishedStatement, CanUseComment, CanUseConcurrently):
    """
    Represents HNSW index
    There is code duplication in this class. This is made specifically for IDE hints

    Refer to:
    https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#hnsw-hierarchical-navigable-small-world-since-150
    """

    def __init__(self, statement: Statement, dimension_number: int):
        super().__init__(statement)
        self._num = dimension_number
        self._dist = None
        self._efc = None
        self._type = None
        self._m = None

    def type_f64(self) -> "HNSW":
        """
        The TYPE clause is optional and can be used to specify the data type of the vector.
        Represents 64-bit floating-point numbers (double precision floating-point numbers).

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#types

        :return: HNSW object
        """
        self._type = "F64"
        return self

    def type_f32(self) -> "HNSW":
        """
        The TYPE clause is optional and can be used to specify the data type of the vector.
        Represents 32-bit floating-point numbers (single precision floating-point numbers).

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#types

        :return: HNSW object
        """
        self._type = "F32"
        return self

    def type_i64(self) -> "HNSW":
        """
        The TYPE clause is optional and can be used to specify the data type of the vector.
        Represents 64-bit signed integers.

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#types

        :return: HNSW object
        """
        self._type = "I64"
        return self

    def type_i32(self) -> "HNSW":
        """
        The TYPE clause is optional and can be used to specify the data type of the vector.
        Represents 32-bit signed integers.

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#types

        :return: HNSW object
        """
        self._type = "I32"
        return self

    def type_i16(self) -> "HNSW":
        """
        The TYPE clause is optional and can be used to specify the data type of the vector.
        Represents 16-bit signed integers.

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#types

        :return: HNSW object
        """
        self._type = "I16"
        return self

    def distance_euclidean(self) -> "HNSW":
        """
        Make index to use Euclidean distance
        When no function is specified, SurrealDB chooses Euclidean for the default distance function

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/functions/vector#vectordistanceeuclidean

        :return: HNSW object
        """
        self._dist = "EUCLIDEAN"
        return self

    def distance_manhattan(self) -> "HNSW":
        """
        Make index to use Manhattan distance
        When no function is specified, SurrealDB chooses Euclidean for the default distance function

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/functions/vector#vectordistancemanhattan

        :return: HNSW object
        """
        self._dist = "MANHATTAN"
        return self

    def distance_cosine(self) -> "HNSW":
        """
        Make index to use Cosine distance
        When no function is specified, SurrealDB chooses Euclidean for the default distance function

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/functions/vector#vectorsimilaritycosine

        :return: HNSW object
        """
        self._dist = "COSINE"
        return self

    def distance_minkowski(self) -> "HNSW":
        """
        Make index to use Minkowski distance
        When no function is specified, SurrealDB chooses Euclidean for the default distance function

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/functions/vector#vectordistanceminkowski

        :return: HNSW object
        """
        self._dist = "MINKOWSKI"
        return self

    def efc(self, value: int) -> "HNSW":
        """
        Exploration factor during construction. This parameter determines the size of the dynamic list for the nearest
        neighbor candidates during the graph construction phase. A larger efConstruction value leads to a more thorough
        construction, improving the quality and accuracy of the search but increasing construction time.
        The default value is 150 (on SurrealDB side)

        :param value: integer value for EFC
        :return: HNSW object
        """
        self._efc = value
        return self

    def max_connections(self, value: int) -> "HNSW":
        """
        Max Connections per Element. Defines the maximum number of bidirectional links (neighbors) per node in each
        layer of the graph, except for the lowest layer. This parameter controls the connectivity and overall structure
        of the network. Higher values of MM generally improve search accuracy but increase memory usage and
        construction time.
        The default value is 12 (on the SurrealDB side)

        :param value: integer value for MM
        :return: HNSW object
        """
        self._m = value
        return self

    def _clean_str(self):
        type_ = f' TYPE {self._type}' if self._type else ''
        dist = f' DIST {self._dist}' if self._dist else ''
        efc = f' EFC {self._efc}' if self._efc else ''
        m = f' M {self._m}' if self._m else ''
        return f"{self._statement._clean_str()} HNSW DIMENSION {self._num}{type_}{dist}{efc}{m}"


class CanUseIndexTypes:
    """
    Represents interface for index types
    """

    def unique(self) -> Unique:
        """
        Creates unique index

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#unique-index

        :return: Unique object
        """
        return Unique(self)

    def search_analyzer(self, name: str) -> SearchAnalyzer:
        """
        Creates full-text search index

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#full-text-search-index

        :param name: analyzer name
        :return: SearchAnalyzer object
        """
        return SearchAnalyzer(self, name)

    def mtree(self, num_dimension: int) -> MTree:
        """
        Creates vector-search index MTREE

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#vector-search-indexes

        :param num_dimension: number of dimensions
        :return: MTree object
        """
        return MTree(self, num_dimension)

    def hnsw(self, num_dimension: int) -> HNSW:
        """
        Creates vector-search index HNSW

        Refer to:
        https://surrealdb.com/docs/surrealdb/surrealql/statements/define/indexes#hnsw-hierarchical-navigable-small-world-since-150

        :param num_dimension: number of dimensions
        :return: HNSW object
        """
        return HNSW(self, num_dimension)



import logging
import warnings
from typing import Optional, Tuple, List, Dict, Union, Any, Callable

from surrealist.connections.connection import Connection
from surrealist.enums import AutoOrNone, Algorithm
from surrealist.errors import SurrealConnectionError
from surrealist.ql.statements import Select, Remove, Live
from surrealist.ql.statements.alter import Alter
from surrealist.ql.statements.define import (DefineEvent, DefineParam, DefineScope,
                                             DefineIndex, DefineToken, DefineTable, DefineField)
from surrealist.ql.statements.define_access import DefineAccessJwt, DefineAccessRecord
from surrealist.ql.statements.define_analyzer import DefineAnalyzer
from surrealist.ql.statements.define_config import DefineConfig
from surrealist.ql.statements.define_user import DefineUser
from surrealist.ql.statements.rebuild_index import RebuildIndex
from surrealist.ql.statements.relate import Relate
from surrealist.ql.statements.returns import Return
from surrealist.ql.statements.statement import Statement
from surrealist.ql.statements.transaction import Transaction
from surrealist.ql.table import Table
from surrealist.result import SurrealResult
from surrealist.surreal import Surreal
from surrealist.utils import DEFAULT_TIMEOUT, NS, DB, AC

logger = logging.getLogger("surrealist.databaseQL")


class Database:
    """
    Represents connected database(in some namespace) to operate on.
    It has features of the database level, including switch to table level. You can use DEFINE/REMOVE, query or
    transactions here, but for simple CRUD you need to switch to table level

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py
    """

    def __init__(self, url: str, namespace: str, database: str, access: Optional[str] = None,
                 credentials: Optional[Tuple[str, str]] = None,
                 use_http: bool = False, timeout: int = DEFAULT_TIMEOUT,
                 active_connection: Optional[Connection] = None):
        """
        Creates a new connection to the database or uses existing connection
        :param url: url of the SurrealDB
        :param namespace: name of the namespace
        :param database: name of the database
        :param access: access method
        :param credentials: pair of username and password
        :param use_http: uses http-transport if True, websockets otherwise
        :param timeout: timeout for the queries
        :param active_connection: existing and active (connected) connection to use, If specified, all other
        parameters are ignored
        """
        if active_connection is None:
            self._namespace = namespace
            self._database = database
            self._access = access
            self._connection = Surreal(url, namespace, database, access=access, credentials=credentials,
                                       use_http=use_http, timeout=timeout).connect()
            logger.info("DatabaseQL is up")
        else:
            self._connection = self._use_connection(active_connection)
            logger.info("DatabaseQL is up, using existing connection")
        self._connected = True

    def _use_connection(self, connection: Connection) -> Connection:
        is_connected = connection.is_connected()
        if not is_connected:
            raise SurrealConnectionError("Cant use connection which is not active")
        db_params = connection._db_params
        if not db_params or db_params.get(NS) is None or db_params.get(DB) is None:
            raise SurrealConnectionError("Can use only database level connections, specify namespace and database")
        ns = db_params[NS]
        db = db_params[DB]
        access = db_params.get(AC)
        self._namespace = ns
        self._database = db
        self._access = access
        return connection

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        self.close()

    def is_connected(self) -> bool:
        """
        Return if a database is connected
        :return: True if connection is active, False otherwise
        """
        return self._connected

    def get_connection(self) -> Connection:
        """
        Returns the underlying connection
        :return: Connection object
        """
        return self._connection

    @property
    def namespace(self) -> str:
        """
        Return name if the namespace
        :return: namespace
        """
        return self._namespace

    @property
    def name(self) -> str:
        """
        Return name of the database
        :return: database name
        """
        return self._database

    def close(self):
        """
        Closes the connection. You cannot and should not use a database object after that
        """
        logger.info("DatabaseQL is closed")
        self._connection.close()
        self._connected = False

    @classmethod
    def from_connection(cls, connection: Connection) -> "Database":
        """
        Builds a database object from an active existing connection
        :param connection: connection to use
        :return: Database object
        """
        return Database("", "", "", None, None, False, 0, connection)

    def tables(self) -> List[str]:
        """
        Return list of the table names at a current database
        :return: string list with the names
        """
        logger.info("Get tables for db %s", self._database)
        return self._connection.db_tables().result

    def info(self) -> Dict:
        """
        Return full info about a database, actually call for "INFO FOR DB;" via query
        :return: a result of the response
        """
        logger.info("Get info for db %s", self._database)
        return self._connection.db_info().result

    def raw_query(self, query: str) -> SurrealResult:
        """
        Execute a raw QL query on a database, you should use it in cases when your query is large or complicated, or it
        is just more appropriate for you.

        :param query: text of the query, it will be sent to SurrealDB as is
        :return: a result of the response
        """
        logger.info("Query for db %s", self._database)
        return self._connection.query(query)

    def returns(self, query: str) -> Return:
        """
        Return result of the query

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/return

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param query: string query to execute
        :return: Return object
        """
        return Return(self._connection, query)

    def __getattr__(self, item) -> Table:
        """
        We actually need this to use dot as the switch to table level as "db.person" instead of "db.table("person")"
        :param item: name of the table to work with
        :return: a table object
        """
        return Table(item, self._connection)

    def table(self, name) -> Table:
        """
        Switch to the Table level(an object) after that you can and should use table operations (CRUD)

        :param name: name of the table to work with
        :return: a table object
        """
        return Table(name, self._connection)

    def transaction(self, statements: List[Statement]) -> Transaction:
        """
        Create a transaction object to generate a query or run

        Refer to: https://docs.surrealdb.com/docs/surrealql/transactions

        :param statements: list of appropriate statements (select, create, delete. etc.)
        :return: Transaction object
        """
        return Transaction(self._connection, statements)

    def select_from(self, select: Select, *args, alias: Optional[Tuple[str, Union[str, Statement]]] = None,
                    value: Optional[str] = None) -> Select:
        """
        Allow selecting from sub-query (select)

        :param select: Select object
        :param args: which fields to select, if no fields or "*" - selects all
        :param alias: pairs of names and values
        :param value: on exists add VALUE statement
        :return: Select object
        """
        return Select(self._connection, select, *args, alias=alias, value=value)

    def define_event(self, name: str, table_name: str, then: Union[str, Statement]) -> DefineEvent:
        """
        Allow defining event on table

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/event

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param name: name for the event
        :param table_name: name of the table
        :param then: action to perform
        :return: DefineEvent object
        """
        return DefineEvent(self._connection, name, table_name, then)

    def remove_event(self, name: str, table_name: str) -> Remove:
        """
        Remove an event linked to table

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/remove

        :param name: name of the event
        :param table_name: name of the table
        :return: Remove object
        """
        return Remove(self._connection, table_name=table_name, type_="EVENT", name=name)

    def define_user(self, user_name: str) -> DefineUser:
        """
        Allow defining user for a current database

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/user

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param user_name: name for the new user
        :return: DefineUser object
        """
        return DefineUser(self._connection, user_name=user_name)

    def remove_user(self, user_name: str) -> Remove:
        """
        Remove user of the database

        :param user_name: name of the user
        :return: Remove object
        """
        return Remove(self._connection, "", type_="USER", name=user_name)

    def define_param(self, name: str, value: Any) -> DefineParam:
        """
        Represents DEFINE PARAM statement

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/param

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param name: name of the parameter
        :param value: value for the parameter
        :return: DefineParam object
        """
        return DefineParam(self._connection, name, value)

    def remove_param(self, name: str) -> Remove:
        """
        Remove parameter of the database

        :param name: name of the parameter
        :return: Remove object
        """
        return Remove(self._connection, "", type_="PARAM", name=name)

    def define_analyzer(self, name: str) -> DefineAnalyzer:
        """
        Represents DEFINE ANALYZER statement

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/analyzer

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param name: name for the analyzer
        :return: DefineAnalyzer object
        """
        return DefineAnalyzer(self._connection, name)

    def remove_analyzer(self, name: str) -> Remove:
        """
        Remove the analyzer

        :param name: name of the analzer
        :return: Remove object
        """
        return Remove(self._connection, '', type_="ANALYZER", name=name)

    def define_scope(self, name: str, duration: str, signup: Union[str, Statement],
                     signin: Union[str, Statement]) -> DefineScope:
        """
        Deprecated since SurrealDB 2.x, use define_access_record instead!

        Represents DEFINE SCOPE statement

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/scope

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param name: name for the new scope
        :param duration: session duration, like 24h
        :param signup: Create statement with string or Statement representation
        :param signin: Select statement with string or Statement representation
        :return: DefineScope object
        """
        msg = "Deprecated since SurrealDB 2.x, use define_access_record instead\n" \
              "Read more here: https://surrealdb.com/docs/surrealql/statements/define/access"
        warnings.warn(msg)
        logger.warning(msg)
        return DefineScope(self._connection, name, duration, signup, signin)

    def remove_scope(self, name: str) -> Remove:
        """
        Do not work since SurrealDB 2.x, use remove_access instead!

        Remove the scope

        :param name: name of the scope
        :return: Remove object
        """
        msg = "Do not work since SurrealDB 2.x, use remove_access instead\n" \
              "Read more here: https://surrealdb.com/docs/surrealql/statements/remove"
        warnings.warn(msg)
        logger.warning(msg)
        return Remove(self._connection, "", type_="SCOPE", name=name)

    def define_index(self, name: str, table_name: str) -> DefineIndex:
        """
        Represents DEFINE INDEX statement

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/indexes

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param name: name for the index
        :param table_name: name of table to work with
        :return: DefineINdex object
        """
        return DefineIndex(self._connection, name, table_name)

    def rebuild_index(self, index_name: str, table_name: str, if_exists: bool = False) -> RebuildIndex:
        """
        Represents REBUILD INDEX object, used to rebuild resources.

        Refer to: https://surrealdb.com/docs/surrealdb/surrealql/statements/rebuild

        :param index_name: name of the index
        :param table_name: name of the table
        :param if_exists: use IF EXISTS statement if True
        :return: RebuildIndex object
        """
        return RebuildIndex(self._connection, index_name=index_name, table_name=table_name, if_exists=if_exists)

    def remove_index(self, name: str, table_name: str) -> Remove:
        """
        Remove index by name on the table

        :param name: name og the index
        :param table_name: name of the table
        :return: Remove object
        """
        return Remove(self._connection, name=name, table_name=table_name, type_="INDEX")

    def define_token(self, name: str, token_type: Algorithm, value: str) -> DefineToken:
        """
        Deprecated since SurrealDB 2.x, use define_access_jwt or define_access_record instead!

        Represents DEFINE TOKEN statement

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/token

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param name: name for the token
        :param token_type: type of the token, for example, Algorithm.RS256
        :param value: value of the token
        :return: DefineToken object
        """
        msg = "Deprecated since SurrealDB 2.x, use define_access_jwt or define_access_record instead\n" \
              "Read more here: https://surrealdb.com/docs/surrealql/statements/define/access"
        warnings.warn(msg)
        logger.warning(msg)
        return DefineToken(self._connection, name, token_type, value)

    def remove_token(self, name: str) -> Remove:
        """
        Do not work since SurrealDB 2.x, use remove_access instead!

        Remove token by name for the database

        :param name: name of the token
        :return: Remove object
        """
        msg = "Do not work since SurrealDB 2.x, use remove_access instead\n" \
              "Read more here: https://surrealdb.com/docs/surrealql/statements/remove"
        warnings.warn(msg)
        logger.warning(msg)
        return Remove(self._connection, "", type_="TOKEN", name=name)

    def define_access_jwt(self, name: str) -> DefineAccessJwt:
        """
        Represents DEFINE ACCESS ... JWT statement.
        Use this method instead of define_token

        Refer to: https://surrealdb.com/docs/surrealql/statements/define/access/jwt

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/define_access.py

        :param name: name for the access
        :return: DefineAccessJwt object
        """
        return DefineAccessJwt(self._connection, name)

    def define_access_record(self, name: str) -> DefineAccessRecord:
        """
        Represents DEFINE ACCESS ... RECORD statement.
        Use this method instead of define_token or define_scope

        Refer to: https://surrealdb.com/docs/surrealql/statements/define/access/record

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/define_access.py

        :param name: name for the access
        :return: DefineAccessRecord object
        """
        return DefineAccessRecord(self._connection, name)

    def remove_access(self, name: str) -> Remove:
        """
        Remove access by name for the database

        :param name: name of the access
        :return: Remove object
        """
        return Remove(self._connection, "", type_="ACCESS", name=name)

    def relate(self, value: str) -> Relate:
        """
        Represents RELATE statement, it should be able to use any statements from documentation

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/relate

        Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_relate_examples.py

        :param value: relate representation, see examples
        :return: Relate object
        """
        return Relate(self._connection, value=value)

    def live_query(self, table_name: str, callback: Callable[[Dict], None], select: Optional[str] = None,
                   use_diff: bool = False) -> Live:
        """
        Represents LIVE statement for a live query

        Example:
        db.live_query("person", func).alias("first_name", "NAME").where("age > 22").run()

        Refer to: https://surrealdb.com/docs/surrealdb/surrealql/statements/live

        Refer to: https://github.com/kotolex/surrealist?tab=readme-ov-file#live-query

        Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_live_examples.py

        :param table_name: name od the table to live select
        :param callback: function to call on live query event, signature is `def callback(arg:Dict) -> None`
        :param select: raw query to insert between LIVE SELECT and FROM {table}, so the result will be
        LIVE SELECT {select} FROM {table_name}.
        If it is provided, other parameters (diff, alias, value) will be ignored
        :param use_diff: return result in DIFF format
        :return: Live object
        """
        return Live(self._connection, table_name, callback, select, use_diff)

    def kill_query(self, live_id: str) -> SurrealResult:
        """
        Represents a KILL statement, for killing a live query by id

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/kill

        :param live_id: id of the query
        :return: result
        """
        return self._connection.kill(live_id)

    def define_table(self, name: str) -> DefineTable:
        """
        Represents DEFINE TABLE statement

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/table

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param name: name of the new table
        :return: DefineTable object
        """
        return DefineTable(self._connection, name)

    def define_field(self, field_name: str, table_name: str) -> DefineField:
        """
        Represents DEFINE FIELD statement

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/field

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/define_field.py

        :param field_name: name of the field
        :param table_name: name of the table
        :return: DefineField object
        """
        return DefineField(self._connection, field_name, table_name)

    def define_config(self, kind: Optional[AutoOrNone] = None) -> DefineConfig:
        """
        Represents DEFINE CONFIG GRAPHQL statement

        Refer to: https://surrealdb.com/docs/surrealql/statements/define/config

        Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/define_config.py

        :param kind: kind of the config
        :return: DefineConfig object
        """
        return DefineConfig(self._connection, kind)

    def remove_field(self, field_name: str, table_name: str) -> Remove:
        """
        Remove field by name for the table

        :param field_name: name of the field
        :param table_name: name of the table
        :return: Remove object
        """
        return Remove(self._connection, table_name, type_="FIELD", name=field_name)

    def remove_table(self, table_name: str) -> Remove:
        """
        Remove table by name for the database

        :param table_name: name of the table to remove
        :return: Remove object
        """
        return Remove(self._connection, table_name, type_="TABLE", name="")

    def alter_table(self, table_name: str) -> Alter:
        """
        Represents ALTER TABLE statement

        Refer to: https://surrealdb.com/docs/surrealdb/surrealql/statements/alter

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_alter_examples.py

        :param table_name: name of the table
        :return: Alter object
        """
        return Alter(self._connection, table_name)

    def run_function(self, function_name: str, version: Optional[str] = None,
                     args: Optional[List] = None) -> SurrealResult:
        """
        This method allows you to execute built-in functions, custom functions, or machine learning models with
        optional arguments

        Refer to: https://surrealdb.com/docs/surrealdb/integration/rpc#run

        Examples:
        connection.run("time::now") # returns current time

        :param function_name: The name of the function or model to execute. Prefix with fn:: for custom functions or
        ml:: for machine learning models.
        :param version: optional parameter, the version of the function or model to execute. When using a machine
        learning model (prefixed with ml::), the version parameter is required.
        :param args: The arguments to pass to the function or model.
        :return: result of request
        """
        return self._connection.run(function_name, version=version, args=args)

    def __repr__(self):
        return f"Database(namespace={self._namespace}, name={self._database}, connected={self.is_connected()}, " \
               f"connection with {self._connection.transport().value} transport)"



import json
from typing import Optional, Union, Dict, List, Any

from surrealist.errors import TooManyNestedLevelsError, ResultHasNoValuesError
from surrealist.utils import OK, ERR, HTTP_OK


class SurrealResult:
    """
    Represents a result of the request both via http or websocket.
    Contains a few helpers to work with results of different kinds: id, ids, get, is_error, is_empty

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/result.py
    """

    def __init__(self, **kwargs):
        """
        Expected fields:
        id - only from websocket requests
        result - error text or any other result of the request
        code - http status code from http requests or error code from websocket
        query - text of the query if available
        status - OK or ERR, if ERR then result contains error text
        time - execution time, only for http requests
        additional_info -all other fields
        """
        self.ws_id: Optional[Union[int, str]] = kwargs.pop("id", None)
        self.result: Optional[Union[str, int, Dict, List]] = kwargs.pop("result", None)
        self.code: Optional[int] = kwargs.pop("code", None)
        self.query: Optional[str] = kwargs.pop("query", None)
        self.status: Optional[str] = kwargs.pop("status", OK)
        self.time: Optional[str] = kwargs.pop("time", None)
        if "information" in kwargs:
            self.result = kwargs.pop("information")
        if "error" in kwargs:
            self.result = kwargs.pop("error")
            self.status = ERR
        if "token" in kwargs:
            self.result = kwargs.pop("token")
        if self.code and self.code != HTTP_OK:
            self.status = ERR
        self.additional_info: Dict = kwargs
        if self.status == ERR and isinstance(self.result, dict) and "code" in self.result and "message" in self.result:
            self.code = self.result["code"]
            self.result = self.result["message"]
        if self.result and isinstance(self.result, str) and "There was a problem with the database:" in self.result:
            self.result = self.result.split(":", 1)[1].strip()

    def count(self) -> int:
        """
        Returns number of records in result.

        :return: int value of records in a result
        """
        if self.is_empty():
            return 0
        if isinstance(self.result, List):
            return len(self.result)
        return 1

    def is_empty(self) -> bool:
        """
        Return is result is empty

        :return: True, if a result is None, empty list, empty dict or empty string, False otherwise
        """
        return self.result in (None, [], {}, '')

    @property
    def id(self) -> str:
        """
        Returns id field from a result.
        Note: this property raise on any problem with getting id, sometimes it will be more effective to use ids instead

        :return: id from the record in result
        :raise ValueError: if a result is empty or no id in it
        """
        count = self.count()
        if count != 1:
            raise ValueError(f"No id at result or more than 1 element, body: {self.result}")
        if not isinstance(self.result, (List, Dict)):
            raise ValueError(f"Cant get id, body: {self.result}")
        value = self.result[0] if isinstance(self.result, List) else self.result
        if not isinstance(value, Dict) or "id" not in value:
            raise ValueError(f"No id field, body: {self.result}")
        return value["id"]

    @property
    def ids(self) -> List:
        """
        Returns list with all found id of records inside.
        Note: this property is never raise an error, instead it returns [] on an empty result or no id.
        Note: non-empty list returns None on each element without id, so [1,2,3] returns [None, None, None]

        :return: list with id of records inside
        """
        result = []
        if isinstance(self.result, Dict):
            if "id" in self.result:
                result.append(self.result["id"])
        if isinstance(self.result, List):
            result = [e.get("id") if isinstance(e, Dict) else None for e in self.result]
        return result

    def first(self):
        """
        Returns the first element of the result if it is a list. Otherwise, returns element itself, except when a result
        is empty

        :return: first element of the result, or result itself
        :raise ResultHasNoValuesError: if result is empty (None, [], {}, '')
        """
        if self.count() < 1:
            raise ResultHasNoValuesError("Cant get first on an empty result")
        if isinstance(self.result, List):
            return self.result[0]
        return self.result

    def last(self):
        """
        Returns the last element of the result if it is a list. Otherwise, returns element itself, except when a result
        is empty

        :return: last element of the result, or result itself
        :raise ResultHasNoValuesError: if result is empty (None, [], {}, '')
        """
        if self.count() < 1:
            raise ResultHasNoValuesError("Cant get first on an empty result")
        if isinstance(self.result, List):
            return self.result[-1]
        return self.result

    def get(self, field_name: str, default: Optional[Any] = None) -> Any:
        """
        Tries to get value by field name, will work with dict or only dict in a list, in all other cases returns default

        Examples:
        SurrealResult(result={"a":1}).get("a") == 1
        SurrealResult(result={"a":1}).get("b") == None
        SurrealResult(result=[{"a":1}]).get("a") == 1
        SurrealResult(result=[{"a":1}]).get("b") == None
        SurrealResult(result=[{"a":1}, {"a":2}]).get("a") == None # more than one dict in a result
        SurrealResult(result="token").get("a") == None # not a dict


        :param field_name: field name to look in a result
        :param default: value to return, when field not found
        :return: any result
        """
        count = self.count()
        if count != 1 or not isinstance(self.result, (List, Dict)):
            return default
        value = self.result[0] if isinstance(self.result, List) else self.result
        if not isinstance(value, Dict) or field_name not in value:
            return default
        return value[field_name]

    def is_error(self):
        """
        Method to know the error was returned

        :return: True if response contains error, False otherwise
        """
        return self.status != OK

    def to_dict(self) -> Dict:
        """
        Return all data as dict
        :return: dict with all fields
        """
        return {"ws_id": self.ws_id, "status": self.status, "result": self.result, "query": self.query,
                "code": self.code, "time": self.time, "additional_info": self.additional_info}

    def __repr__(self):
        return f"SurrealResult(id={self.ws_id}, status={self.status}, result={self.result}, query={self.query}, " \
               f"code={self.code}, time={self.time}, additional_info={self.additional_info})"

    def __eq__(self, other):
        if other is None or not isinstance(other, SurrealResult):
            return False
        return self.__dict__ == other.__dict__

    def __hash__(self):
        return hash((self.ws_id, self.result, self.status, self.time, self.code, self.query))


def to_result(content: Union[str, Dict, List]) -> SurrealResult:
    """
    Converts str or dict response of SurrealDB to a common object for convenient use

    :param content: response from SurrealDB
    :return: Result object
    """
    if isinstance(content, str):
        try:
            content = json.loads(content)
        except RecursionError as exc:
            raise TooManyNestedLevelsError("Cant serialize object, too many nested levels") from exc
    if isinstance(content, List):
        if len(content) == 1:
            return SurrealResult(**content[0])
        return SurrealResult(result=[SurrealResult(**e) for e in content])
    if _is_result_inside(content):
        res = SurrealResult(**content["result"][0])
        if "id" in content:
            res.ws_id = content["id"]
        return res
    return SurrealResult(**content)


def _is_result_inside(a_dict) -> bool:
    """
    Helper predicate for deep nested objects
    """
    return len(a_dict) in (1, 2) and "result" in a_dict and isinstance(a_dict["result"], List) \
        and len(a_dict["result"]) == 1 and set(a_dict["result"][0].keys()) == {"time", "status", "result"}



from typing import List

from surrealist.connections import Connection
from surrealist.ql.statements.statement import Statement
from surrealist.utils import OK


class Transaction(Statement):
    """
    Represents TRANSACTION operation, it generates and can run a transaction query

    Refer to: https://docs.surrealdb.com/docs/surrealql/transactions

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/transaction.py

    """

    def __init__(self, connection: Connection, actions: List[Statement]):
        super().__init__(connection)
        self._actions = actions

    def validate(self) -> List[str]:
        return [OK]

    def _clean_str(self) -> str:
        text = "BEGIN TRANSACTION;"
        between = "\n".join(statement.to_str() for statement in self._actions)
        end = "COMMIT TRANSACTION"
        return f"{text}\n\n{between}\n\n{end}"



from .create import Create
from .define import DefineEvent, DefineParam
from .define_analyzer import DefineAnalyzer
from .define_user import DefineUser
from .delete import Delete
from .insert import Insert
from .live import Live
from .rebuild_index import RebuildIndex
from .remove import Remove
from .select import Select
from .show import Show
from .transaction import Transaction
from .update import Update

__all__ = ("Create", "Delete", "Insert", "Live", "Remove", "Select", "Show", "Update", "Transaction", "DefineParam",
           "DefineUser", "DefineEvent", "RebuildIndex", "DefineAnalyzer")



from enum import Enum, auto


class Transport(Enum):
    """
    Represents different types of transport for connection
    """
    HTTP = "http"
    WEBSOCKET = "websocket"


class Algorithm(Enum):
    """
    Represents different types of cryptographic algorithms to use for token verification
    Refer to: https://github.com/surrealdb/surrealdb/blob/main/core/src/iam/verify.rs#L17-L72
    """
    HS256 = auto()
    HS384 = auto()
    HS512 = auto()
    EDDSA = auto()
    ES256 = auto()
    ES384 = auto()
    ES512 = auto()
    PS256 = auto()
    PS384 = auto()
    PS512 = auto()
    RS256 = auto()
    RS384 = auto()
    RS512 = auto()

class AutoOrNone(Enum):
    """
    Represents the auto or none option for DEFINE CONFIG
    """
    AUTO = auto()
    NONE = auto()



import json
import threading
import time
from json import JSONDecodeError
from logging import getLogger
from queue import Queue, Empty
from typing import Dict, Callable, Optional

import websocket

from surrealist.errors import WebSocketConnectionClosedError, TooManyNestedLevelsError
from surrealist.result import to_result, SurrealResult
from surrealist.utils import DEFAULT_TIMEOUT, get_uuid, mask_pass

logger = getLogger("surrealist.clients.websocket")


class WebSocketClient:
    """
    Synchronous thread-safe client to work with websockets, always wait a response for message, comparing by id (uuid).
    Every client creates at least two threads (in and out)
    """

    def __init__(self, base_url: str, timeout: int = DEFAULT_TIMEOUT):
        self._ws = None
        self._connected = None
        self._timeout = timeout
        self._base_url = base_url
        thread = threading.Thread(target=self.run, daemon=True)
        thread.start()
        logger.debug("Connecting to %s", base_url)
        self._raise_on_wait(lambda: self._connected is True, timeout=timeout,
                            error_text=f"Not connected to {self._base_url}")
        self._callbacks = {}
        self._messages: Dict[str, Queue] = {}
        logger.debug("Connected to %s, timeout is %s seconds", base_url, timeout)

    def on_message(self, _ws, message: str):
        """
        Called on a message received from the websocket connection.

        :param _ws: connection object
        :param message: string message
        """
        logger.debug("Get message %s", message)
        try:
            mess = json.loads(message)
        except JSONDecodeError as je:
            # Should never happen, all messages via json
            logger.error("Got non-json response %s", message, exc_info=True)
            raise ValueError(f"Got non-json response! {message}") from je
        except RecursionError as e:
            logger.error("Cant deserialize object, too many nested levels")
            raise TooManyNestedLevelsError("Cant serialize object, too many nested levels") from e
        if "id" in mess:
            id_ = mess["id"]
            self._messages[id_].put_nowait(mess)
        else:
            # no id at top level = live query received
            if 'result' in mess:
                live_id = mess['result']['id']
                callback = self._callbacks.get(live_id)
                if callback:
                    logger.debug("Use callback for %s", live_id)
                    callback(mess)
                else:
                    logger.warning("Got a message, but no callback to work with. Message: %s", mess)
            else:
                logger.warning("Got an unexpected message without id and result: %s", mess)

    def on_error(self, _ws, err: Exception):
        """
        Callback on getting any errors with sockets
        """
        logger.error("Websocket connection gets an error %s", err)

    def is_connected(self) -> bool:
        """
        Shows is a websocket client is connected to SurrealDB

        :return: True if connected, False otherwise
        """
        return bool(self._connected)

    def on_open(self, _ws):
        """
        Callback on establishing new connection
        """
        self._connected = True

    def on_close(self, *_ignore):
        """
        Callback on closing websocket connection
        """
        self._connected = False
        logger.debug("Close connection to %s", self._base_url)

    def run(self):
        """
        Constantly waiting for messages on websocket connection, runs in separate thread

        :return: None
        """
        self._ws = websocket.WebSocketApp(self._base_url, header={'sec-websocket-protocol': 'json'},
                                          on_open=self.on_open, on_message=self.on_message,
                                          on_error=self.on_error, on_close=self.on_close)
        self._ws.run_forever(skip_utf8_validation=True)  # works faster

    def send(self, data: Dict, callback: Optional[Callable] = None) -> SurrealResult:
        """
        Method to send messages to SurrealDB, blocks until gets a response or raise on timeout

        :param data: dict with request parameters
        :param callback: function to call on a live query, it is set only for a live method
        :return: result of the request
        :raise TimeoutError: if no response and time is over
        :raise WebSocketConnectionClosed: if the connection was closed while waiting
        """
        id_ = get_uuid()
        data = {"id": id_, **data}
        to_send = data if "additional" not in data else {k: v for k, v in data.items() if k != "additional"}
        try:
            data_string = json.dumps(to_send, ensure_ascii=False)
        except RecursionError as e:
            logger.error("Cant serialize object, too many nested levels")
            raise TooManyNestedLevelsError("Cant serialize object, too many nested levels") from e
        logger.debug("Send data: %s", mask_pass(data_string))
        self._messages[id_] = Queue(maxsize=1)
        self._ws.send(data_string)
        res = self._get_by_id(id_)
        if data['method'] in ('live', 'kill') or "additional" in data:
            if 'error' not in res:
                # now we know live or kill was successful, so now we need to manage callbacks
                self._on_success(data, callback, res)
        return to_result(res)

    def _on_success(self, data: Dict, callback: Callable, result: Dict):
        if data['method'] == 'kill':
            logger.debug("Delete callback for %s", data['params'][0])
            self._callbacks[data['params'][0]] = None
        else:
            # custom query returns nested result
            key = result['result'] if data['method'] == 'live' else result['result'][0]['result']
            logger.debug("Set callback for %s", result['result'])
            self._callbacks[key] = callback

    def _get_by_id(self, id_) -> Dict:
        try:
            result = self._messages[id_].get(timeout=self._timeout)
        except Empty as exc:
            raise TimeoutError(f"Time exceeded: {self._timeout} seconds, no response received") from exc
        finally:
            del self._messages[id_]
        return result

    def _wait_until(self, predicate, timeout, period=0.0005):
        must_end = time.time() + timeout
        while time.time() < must_end:
            if self._connected is False:
                return False, "CLOSED"
            if predicate():
                return True, None
            time.sleep(period)
        return False, "TIME"

    def _raise_on_wait(self, predicate, timeout, error_text):
        """
        Method to wait some condition or raise on timeout

        :param predicate: function to call and check condition
        :param timeout: time in seconds to wait until a condition
        :param error_text: custom error message on fail
        :raise TimeoutError: if condition not met until time
        :raise WebSocketConnectionClosed: if connection was closed while waiting
        """
        result = self._wait_until(predicate, timeout)
        if result == (False, "TIME"):
            logger.error("Time exceeded: %s seconds. Error: %s", timeout, error_text)
            raise TimeoutError(f"Time exceeded: {timeout} seconds. Error: {error_text}")
        if result == (False, "CLOSED"):
            logger.error("Connection %s closed while a client waits on it", self._base_url)
            raise WebSocketConnectionClosedError("Connection closed while a client waits on it")

    def close(self):
        """
        Close websocket client and close websocket connection, you cannot use this object after close
        """
        self._connected = False
        self._ws.close()
        del self._ws
        self._messages.clear()
        self._callbacks.clear()
        logger.debug("Client is closed connection to %s", self._base_url)



from typing import List, Callable, Optional

from surrealist.connections import Connection
from surrealist.ql.statements.live_statements import LiveUseWhere
from surrealist.ql.statements.statement import Statement
from surrealist.result import SurrealResult
from surrealist.utils import OK


class Live(Statement, LiveUseWhere):
    """
    Represents LIVE SELECT statement, it should be able to use any statements from documentation

    Refer to: https://surrealdb.com/docs/surrealdb/surrealql/statements/live

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_live_examples.py

    LIVE SELECT
    [
        [ VALUE ] @fields [ AS @alias ]
        | DIFF
    ]
    FROM @targets
    [ WHERE @conditions ]
    [ FETCH @fields ... ];
    """

    def __init__(self, connection: Connection, table_name: str, callback: Callable, select: Optional[str] = None,
                 use_diff: bool = False):
        """
        Constructor for LIVE statement

        Refer to: https://surrealdb.com/docs/surrealdb/surrealql/statements/live

        :param connection: connection to use
        :param table_name: name of the table to observe
        :param callback: function to call on incoming events
        :param select: raw query to insert between LIVE SELECT and FROM {table}, so the result will be
        LIVE SELECT {select} FROM {table_name}.
        If it is provided, other parameters (diff, alias, value) will be ignored
        :param use_diff: True if you want to get only DIFF info on table events, False for a standard results
        """
        super().__init__(connection)
        self._table_name = table_name
        self._alias = None
        self._diff = use_diff
        self._callback = callback
        self._select = select
        self._value = None

    def value(self, field_name: str) -> "Live":
        """
        Add VALUE statement

        :param field_name: field to use
        :return: Live object
        """
        self._value = field_name
        return self

    def alias(self, value_name: str, alias: str) -> "Live":
        """
        Add AS statement
        @fields [ AS @alias ]

        :param value_name: name of the field
        :param alias: name of the alias
        :return: Live object
        """
        self._alias = (value_name, alias)
        self._diff = False
        return self

    def validate(self) -> List[str]:
        if self._select and any([self._value, self._diff, self._alias]):
            return ["If select is provided, value, diff and alias parameters will be ignored"]
        if self._value and any([self._diff, self._alias]):
            return ["Using Value with DIFF or alias parameters"]
        if self._alias and self._diff:
            return ["Using DIFF with alias parameter"]
        return [OK]

    def run(self) -> SurrealResult:
        return self._drill(self.to_str())

    def _drill(self, query):
        return self._connection.custom_live(query, self._callback)

    def _clean_str(self):
        what = "*"
        if self._value:
            what = f"VALUE {self._value}"
        if self._diff:
            what = "DIFF"
        elif self._alias:
            what = f"{self._alias[0]} AS {self._alias[1]}"
        if self._select:
            what = self._select
        return f"LIVE SELECT {what} FROM {self._table_name}"
